def rotations(tile, and_reflections=True):
    if and_reflections:
        return rotate_and_reflect(tile)
    else:
        return rotate(tile)


ROTATIONS = [[[1, 0], [0, 1]], [[0, -1], [1, 0]], [[-1, 0], [0, -1]], [[0, 1], [-1, 0]]]


ROTATIONS_AND_REFLECTIONS = [
    [[1, 0], [0, 1]],
    [[0, -1], [1, 0]],
    [[-1, 0], [0, -1]],
    [[0, 1], [-1, 0]],
    [[-1, 0], [0, 1]],
    [[0, -1], [-1, 0]],
    [[1, 0], [0, -1]],
    [[0, 1], [1, 0]],
]


def rotate_and_reflect(tile):
    return unique_after_transform(tile, ROTATIONS_AND_REFLECTIONS)


def rotate(tile):
    return unique_after_transform(tile, ROTATIONS)


def unique_after_transform(tile, transforms):
    s = set()
    for m in transforms:
        rotated = [
            (m[0][0] * t[0] + m[0][1] * t[1], m[1][0] * t[0] + m[1][1] * t[1])
            for t in tile
        ]
        mx = min(s[0] for s in rotated)
        my = min(s[1] for s in rotated)
        shifted = [(t[0] - mx, t[1] - my) for t in rotated]
        key = tuple(sorted(shifted))
        if key not in s:
            yield shifted
            s.add(key)



from pretty_poly import make_ascii

from .jupyter import solution_to_png


class Solution(object):
    def __init__(self, tiling, board):
        self.tiling = tiling
        self.board = board

    def display(self):
        return make_ascii(self.tiling)

    def python(self):
        return "\n".join(self._gen_python())

    def _gen_python(self):
        yield "TILING = %s" % (repr(self.tiling),)

    def _repr_png_(self):
        return solution_to_png(self)



from polyomino.board import Rectangle
from polyomino.error import CantPlaceSinglePiece


def rectangles(size):
    """
        >>> list(rectangles(20):
        [(1, 20), (2, 10), (4, 5)]
        >>> list(rectangles(16):
        [(1, 16), (2, 8), (4, 6)]
    """
    for i in range(1, size):
        if size % i == 0:
            yield i, size // i


def has_order_greater_than(tile, k):
    n = len(tile)

    for i in range(1, k + 1):
        for x, y in rectangles(n * i):
            board = Rectangle(x, y)
            problem = board.tile_with_many(tile)
            try:
                solution = problem.solve()
                if solution:
                    return False, i, solution.tiling
            except CantPlaceSinglePiece:
                pass

    return True, None, None



import io

from pretty_poly.image import make_colored_blocks, make_lines
from pretty_poly.png import _make_png_writer


def solution_to_png(solution):
    data, palette = make_colored_blocks(solution.tiling)
    writer = _make_png_writer(data, palette)
    f = io.BytesIO()
    writer.write(f, data)
    return f.getvalue()


def board_to_png(board):
    data, palette = make_lines([board.squares])
    writer = _make_png_writer(data, palette)
    f = io.BytesIO()
    writer.write(f, data)
    return f.getvalue()



import numpy as np

from pretty_poly import make_ascii

from .jupyter import board_to_png
from .problem import TilingProblem
from .tileset import exactly, many
from .transform import rotations


class Shape(object):
    bit_masks = None

    def is_contained(self, tile):
        return all(self.is_in(sq) for sq in tile)

    @property
    def count(self):
        return len(self.squares)

    def positions(self, tile, with_reflections=False):
        for sq in self.squares:
            for rotated in rotations(tile, with_reflections):
                reference = rotated[0]
                translated = [
                    (x - reference[0] + sq[0], y - reference[1] + sq[1])
                    for x, y in rotated
                ]
                if self.is_contained(translated):
                    yield translated

    def bit_vector(self, tile):
        if not tile:
            return self.bit_mask(None)
        return np.sum([self.bit_mask(sq) for sq in tile], axis=0)

    def bit_mask(self, sq):
        if not self.bit_masks:
            self.bit_masks = dict()
        if sq not in self.bit_masks:
            self.bit_masks[sq] = np.array(
                [s == sq for s in self.squares], dtype=np.int32
            )
        return self.bit_masks[sq]

    def remove(self, square):
        if square not in self.squares:
            raise Exception(
                "Tried to remove a square %s which was not present in the board"
                % (square,)
            )
        return Irregular([sq for sq in self.squares if sq != square])

    def remove_all(self, squares):
        for square in squares:
            if square not in self.squares:
                raise Exception(
                    "Tried to remove a square %s which was not present in the board"
                    % (square,)
                )
        return Irregular([sq for sq in self.squares if sq not in squares])

    def tile_with(self, tiles):
        return self.tile_with_set(exactly(tiles))

    def tile_with_many(self, tile):
        return self.tile_with_set(many(tile))

    def tile_with_set(self, tileset):
        tileset.check(self)
        return TilingProblem(self, tileset)

    def display(self):
        return make_ascii([self.squares])

    def format_tiling(self, pieces):
        squares_in_pieces = (sq for piece in pieces for sq in piece)
        remainder = set(self.squares) - set(squares_in_pieces)
        return make_ascii(pieces + [remainder])

    def _repr_png_(self):
        return board_to_png(self)


class Irregular(Shape):
    _adjusted = None

    def __init__(self, squares):
        self._squares = list(squares)
        self.min_x = min(x for x, y in self.squares)
        self.min_y = min(y for x, y in self.squares)
        self.max_x = max(x for x, y in self.squares)
        self.max_y = max(y for x, y in self.squares)

    @property
    def squares(self):
        return self._squares

    @property
    def adjusted(self):
        if self._adjusted is None:
            self._adjusted = [(x - self.min_x, y - self.min_y) for x, y in self.squares]
        return self._adjusted

    def is_in(self, square):
        return square in self._squares


class DeletedRectangle(Irregular):
    def __init__(self, whole, deleted):
        self.whole = whole
        self.deleted = deleted
        super().__init__([sq for sq in self.whole.squares if not sq in self.deleted])

    def remove(self, square):
        if square not in self.squares:
            raise Exception(
                "Tried to remove a square %s which was not present in the board"
                % (square,)
            )
        return DeletedRectangle(self.whole, self.deleted + [square])

    def remove_all(self, squares):
        for square in squares:
            if square not in self.squares:
                raise Exception(
                    "Tried to remove a square %s which was not present in the board"
                    % (square,)
                )
        return DeletedRectangle(self.whole, self.deleted + squares)

    @property
    def interior_deleted(self):
        return (
            (x, y)
            for x, y in self.deleted
            if x > self.min_x and x < self.max_x and y > self.min_y and y < self.max_y
        )

    def format_row_sides(self, row, row_y):
        line = list(super().format_row_sides(row))
        for x, y in self.interior_deleted:
            if y == row_y:
                line[2 * x + 1] = "X"
        return "".join(line)

    def format_tiling_lines(self, h, v):
        for i in range(self.min_y, self.max_y + 1):
            yield self.format_row_upper(h[i])
            yield self.format_row_sides(v[i - self.min_y], i)
        yield self.format_row_upper(h[self.max_y - self.min_y + 1])


class Rectangle(Shape):
    _squares_cache = None

    def __init__(self, x, y):
        self.x = x
        self.y = y

    @property
    def squares(self):
        if self._squares_cache is None:
            self._squares_cache = list(self.gen_squares())
        return self._squares_cache

    def gen_squares(self):
        for i in range(0, self.x):
            for j in range(0, self.y):
                yield (i, j)

    def is_in(self, square):
        return 0 <= square[0] < self.x and 0 <= square[1] < self.y

    def format_row_sides(self, row):
        return " ".join("|" if r else " " for r in row)

    def format_row_upper(self, row):
        return "+" + "+".join("-" if r else " " for r in row) + "+"

    def format_tiling_lines(self, h, v):
        for i in range(0, self.y):
            yield self.format_row_upper(h[i])
            yield self.format_row_sides(v[i])
        yield self.format_row_upper(h[self.y])

    def calculate_tiling(self, tiling):
        h = [[True] * self.x for i in range(0, self.y + 1)]
        v = [[True] * (self.x + 1) for i in range(0, self.y)]

        for tile in tiling:
            for sq_a in tile:
                for sq_b in tile:
                    a, b = sorted([sq_a, sq_b])
                    if (a[1] == b[1]) and (a[0] + 1 == b[0]):
                        v[a[1]][b[0]] = False
                        pass
                    if (a[0] == b[0]) and (a[1] + 1 == b[1]):
                        h[b[1]][a[0]] = False

        return h, v

    def format_tiling(self, tiling):
        h, v = self.calculate_tiling(tiling)
        return "\n".join(self.format_tiling_lines(h, v))

    def remove(self, square):
        if square not in self.squares:
            raise Exception(
                "Tried to remove a square %s which was not present in the board"
                % (square,)
            )
        return DeletedRectangle(self, [square])

    def remove_all(self, squares):
        for square in squares:
            if square not in self.squares:
                raise Exception(
                    "Tried to remove a square %s which was not present in the board"
                    % (square,)
                )
        return DeletedRectangle(self, squares)


class Chessboard(Rectangle):
    x = 8
    y = 8

    def __init__(self):
        pass



class PolyominoError(Exception):
    pass


class CoverWithWrongSize(PolyominoError):
    def __init__(self, board_size, cover_size):
        self.board_size = board_size
        self.cover_size = cover_size


class CoverWithWrongModulus(PolyominoError):
    def __init__(self, board_size, fixed_size, gcd):
        self.board_size = board_size
        self.fixed_size = fixed_size
        self.gcd = gcd


class CantPlaceSinglePiece(PolyominoError):
    def __init__(self, board, tileset):
        self.board = board
        self.tileset = tileset



import numpy as np

from .error import CoverWithWrongSize, CoverWithWrongModulus
from .utils import gcd_list


class Tileset(object):
    def __init__(self, mandatory, optional, filler, reflections=False):
        self.mandatory = mandatory  # exactly 1
        self.optional = optional  # either 1 or 0
        self.filler = filler  # any nonnegative number
        self.reflections = reflections

    @property
    def fixed_total(self):
        return sum(len(tile) for tile in self.mandatory)

    @property
    def optional_total(self):
        return sum(len(tile) for tile in self.optional)

    @property
    def max_total(self):
        return self.fixed_total + self.optional_total

    @property
    def selector_size(self):
        return len(self.mandatory) + len(self.optional)

    @property
    def flex_gcd(self):
        return gcd_list(
            [len(tile) for tile in self.optional] + [len(tile) for tile in self.filler]
        )

    def check(self, board):
        n = board.count
        if self.fixed_total > n:
            raise CoverWithWrongSize(n, self.fixed_total)
        if not self.filler:
            if self.max_total < n:
                raise CoverWithWrongSize(n, self.fixed_total)

        r = n - self.fixed_total
        g = self.flex_gcd
        if r != 0 and r % g != 0:
            raise CoverWithWrongModulus(n, self.fixed_total, g)

    def selector_vector(self, i):
        return np.array([i == j for j in range(0, self.selector_size)])

    def empty_vector(self):
        return np.array([False] * self.selector_size)

    def vectors(self):
        i = 0
        for tile in self.mandatory:
            selector = self.selector_vector(i)
            yield tile, selector, False
            i = i + 1
        for tile in self.optional:
            selector = self.selector_vector(i)
            yield tile, selector, True
            i = i + 1
        for tile in self.filler:
            selector = self.empty_vector()
            yield tile, selector, False

    def and_repeated_exactly(self, count, tile):
        mandatory = self.mandatory + [tile] * count
        return Tileset(mandatory, self.optional, self.filler)

    def and_one(self, tile):
        return self.and_repeated_exactly(1, tile)

    def with_reflections(self):
        return Tileset(self.mandatory, self.optional, self.filler, True)


def many(tile):
    return Tileset([], [], [tile])


def exactly(tiles):
    return Tileset(tiles, [], [])


def repeated_exactly(count, tile):
    return Tileset([tile] * count, [], [])


def any_number_of(tiles):
    return Tileset([], [], tiles)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['polyomino']

package_data = \
{'': ['*']}

install_requires = \
['exact-cover==1.3.0', 'pretty-poly==0.5.1']

entry_points = \
{'console_scripts': ['doctest = run_tests:run_doctests',
                     'test = run_tests:run_tests']}

setup_kwargs = {
    'name': 'polyomino',
    'version': '0.7.1a0',
    'description': 'Solve polyomino tiling problems.',
    'long_description': "# POLYOMINO - a Python package for polyomino tiling problems\n[![PyPI version](https://badge.fury.io/py/polyomino.svg)](https://badge.fury.io/py/polyomino)\n![Deploy wheels to pypi](https://github.com/jwg4/polyomino/workflows/Deploy%20wheels%20to%20pypi/badge.svg)\n![Run Python tests](https://github.com/jwg4/polyomino/workflows/Run%20Python%20tests/badge.svg)\n\n\nThis is a package for manipulating polyominos and in particular, solving tiling problems. It uses the 'exact-cover' python package as the main engine for solving cover problems.\n\nTo solve a tiling problem, you need to create a 'board', the set of squares to be covered, and a 'tileset', the collection of polyominos which can be used. There are examples of the syntax to do this in examples/fluid.md The example file examples/gardner.md uses the package to solve a number of problems from the chapter on polyominos from Martin Gardner's book 'Mathematical Puzzles and Diversions'.\n\nThe folder examples/ also has a couple of Jupyter notebooks: these show how you can work with basic boards and solutions, and how they can be displayed in a notebook. The notebook display should work out-of-the-box.\n\nNote that each tile can play one of several roles in a tiling problem. It could be a tile which can only appear once, as in problems like covering a chessboard with one copy of each pentomino and a square tetromino. It could be a tile which can be used an arbitrary number of times. Problems like this are often simply to cover a shape completely with copies of a single polyomino. Finally, it could be used either once or not at all. When constructing a tileset, it is possible to include tiles in any of these three classes.\n\n## Design\nBoth polyominos and boards are represented internally as lists of integer tuples (x, y). There are constants defined for all polyominos up to pentominos in polyomino.constant\n\nThe search algorithm used for searching for tilings is 'Algorithm X', also known as 'Dancing Links' from the famous paper by Knuth (https://arxiv.org/abs/cs/0011047). The paper suggests polyomino tilings and related problems as examples of exact cover problems which can be solved by the algorithm. There are more details about how the algorithm works on the homepage of the exact-cover package, https://github.com/jwg4/exact_cover\n\nIn reducing a tiling problem to an exact cover problem, each square of the shape to be covered becomes a column of the problem. Each placement of a particular tile in a particular place is a row. A row has 1s in columns corresponding to all the squares which are covered by that shape in that location.\nThere can be additional columns which correspond to the use of a piece, if that piece can only be used once. Problems where each piece can be used an arbitrary number of times, do not have such columns.\n\nIn the case where all tiles must be used exactly once, we have as many columns as the number of squares to be covered plus the number of tiles to be used. As Knuth points out, when the algorithm looks for a column which can be covered in one way, this means it is finding either a square which needs a particular piece to cover it, or for a piece which can only be fitted on the board in one way. Both deductive strategies should be used to narrow down efficiently.\n\nA heuristic for searches involving a limited number of monominos (single squares) can be configured. In the future we hope to implement further search heuristic options, as well as alternative algorithms for special cases like domino tiling.\n\n## Style note\nWe (try to) spell 'polyominos' without an 'e' everywhere, by analogy with 'dominos'.\n",
    'author': 'Jack Grahl',
    'author_email': 'jack.grahl@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



import numpy as np

from exact_cover import get_exact_cover
from exact_cover.error import NoSolution

from .error import CantPlaceSinglePiece
from .solution import Solution


def tiling_to_problem(tileset, shape):
    key = []
    data = []
    for tile, selector, optional in tileset.vectors():
        for translated in shape.positions(tile, tileset.reflections):
            key.append(translated)
            vector = np.concatenate([selector, shape.bit_vector(translated)])
            data.append(vector)
        if optional:
            key.append([])
            vector = np.concatenate([selector, shape.bit_vector([])])
            data.append(vector)
    return key, data


def tiling_to_array(tileset, shape):
    key, data = tiling_to_problem(tileset, shape)
    array = np.array(data, dtype=np.int32)
    return key, array


def tiling_to_array_order_biggest(tileset, shape):
    key, array = tiling_to_array(tileset, shape)
    row_sums = array.sum(axis=1) * -1
    sort_args = np.argsort(row_sums)
    sorted_array = array[sort_args, :]
    sorted_key = [key[i] for i in sort_args]
    return sorted_key, sorted_array


class TilingProblem(object):
    biggest_pieces_first = False
    _name = ""
    array = None
    key = None

    def __init__(self, board, tileset):
        self.board = board
        self.tileset = tileset

    def make_problem(self):
        if self.array is not None and self.key is not None:
            return
        if self.biggest_pieces_first:
            self.key, self.array = tiling_to_array_order_biggest(
                self.tileset, self.board
            )
        else:
            self.key, self.array = tiling_to_array(self.tileset, self.board)
        if self.array.ndim < 2:
            raise CantPlaceSinglePiece(self.board, self.tileset)

    def solve(self):
        self.make_problem()
        try:
            solution = get_exact_cover(self.array)
        except NoSolution:
            return None
        tiling = [self.key[s] for s in solution]
        if tiling:
            return Solution(tiling, self.board)

    def with_heuristics(self):
        self.biggest_pieces_first = True
        return self

    def output_array(self, filename):
        self.make_problem()
        np.savetxt(filename, self.array, fmt="%d", header=self.header)

    def set_name(self, name):
        self._name = name

    @property
    def size(self):
        self.make_problem()
        return self.array.shape

    @property
    def name(self):
        if self._name:
            return self._name
        else:
            return "Tiling problem"

    @property
    def header(self):
        x, y = self.size
        return "%s, %d x %d" % (self.name, x, y)
